import logging
import os
import shutil
from pathlib import Path

import pathvalidate
import posthog
import pytest

from faebryk.libs.util import repo_root as _repo_root
from faebryk.libs.util import robustly_rm_dir

# Disable telemetry for testing
posthog.disabled = True


def pytest_configure(config):
    worker_id = os.environ.get("PYTEST_XDIST_WORKER")
    if worker_id is not None:
        logging.basicConfig(
            format=config.getini("log_file_format"),
            filename=Path("artifacts") / f"tests_{worker_id}.log",
            level=config.getini("log_file_level"),
        )


@pytest.fixture()
def repo_root() -> Path:
    """Fixture providing the repository root path."""
    return _repo_root()


@pytest.fixture()
def setup_project_config(tmp_path):
    from atopile.config import ProjectConfig, ProjectPaths, config

    config.project = ProjectConfig.skeleton(
        entry="", paths=ProjectPaths(build=tmp_path / "build", root=tmp_path)
    )
    yield


@pytest.fixture()
def save_tmp_path_on_failure(tmp_path: Path, request: pytest.FixtureRequest):
    try:
        yield
    except Exception:
        node_name = str(request.node.name)
        safe_node_name = pathvalidate.sanitize_filename(node_name)
        artifact_path = _repo_root() / "artifacts" / safe_node_name
        if artifact_path.exists():
            robustly_rm_dir(artifact_path)
        artifact_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(tmp_path, artifact_path, ignore=shutil.ignore_patterns(".git"))

        raise



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class PowerSwitch(Module):
    """
    A generic module that switches power based on a logic signal, needs specialization

    The logic signal is active high. When left floating, the state is determined by the
    normally_closed parameter.
    """

    def __init__(self, normally_closed: bool) -> None:
        super().__init__()

        self._normally_closed = normally_closed

    logic_in: F.ElectricLogic
    power_in: F.ElectricPower
    switched_power_out: F.ElectricPower

    @L.rt_field
    def switch_power(self):
        return F.can_switch_power_defined(
            self.power_in, self.switched_power_out, self.logic_in
        )

    def __preinit__(self):
        self.switched_power_out.voltage.alias_is(self.power_in.voltage)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class USB_Type_C_Receptacle_14_pin_Vertical(Module):
    """
    14 pin vertical female USB Type-C connector
    918-418K2022Y40000
    """

    # interfaces

    # TODO make arrays?
    cc1: F.Electrical
    cc2: F.Electrical
    shield: F.Electrical
    # power
    vbus: F.ElectricPower
    # diffpairs: p, n
    usb: F.USB2_0

    explicit_part = L.f_field(F.has_explicit_part.by_mfr)(
        "Jing Extension of the Electronic Co.", "918-418K2022Y40000"
    )

    @L.rt_field
    def attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.vbus.lv,
                "2": self.vbus.hv,
                "3": self.usb.usb_if.d.n.line,
                "4": self.usb.usb_if.d.p.line,
                "5": self.cc2,
                "6": self.vbus.hv,
                "7": self.vbus.lv,
                "8": self.vbus.lv,
                "9": self.vbus.hv,
                "10": self.usb.usb_if.d.n.line,
                "11": self.usb.usb_if.d.p.line,
                "12": self.cc1,
                "13": self.vbus.hv,
                "14": self.vbus.lv,
                "0": self.shield,
            }
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import Collection, Iterable

import dash_cytoscape as cyto
from dash import Dash, html
from rich.console import Console
from rich.table import Table

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.graphinterface import GraphInterface
from faebryk.core.link import Link
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.core.parameter import Parameter
from faebryk.core.trait import Trait
from faebryk.exporters.visualize.util import generate_pastel_palette
from faebryk.libs.util import KeyErrorAmbiguous, find_or, typename


# Transformers -------------------------------------------------------------------------
def _gif(gif: GraphInterface):
    return {
        "data": {
            "id": id(gif),
            "label": gif.name,
            "type": typename(gif),
            "parent": id(gif.node),
        }
    }


def _link(source, target, link: Link):
    return {
        "data": {
            "source": id(source),
            "target": id(target),
            "type": typename(link),
        }
    }


_GROUP_TYPES = {
    Parameter: "#FFD9DE",  # Very light pink
    Module: "#E0F0FF",  # Very light blue
    Trait: "#FCFCFF",  # Almost white
    F.Electrical: "#D1F2EB",  # Very soft turquoise
    F.ElectricPower: "#FCF3CF",  # Very light goldenrod
    F.ElectricLogic: "#EBE1F1",  # Very soft lavender
    # Defaults
    ModuleInterface: "#DFFFE4",  # Very light green
    Node: "#FCFCFF",  # Almost white
}


def _group(node: Node):
    try:
        subtype = find_or(_GROUP_TYPES, lambda t: isinstance(node, t), default=Node)
    except KeyErrorAmbiguous as e:
        subtype = e.duplicates[0]

    return {
        "data": {
            "id": id(node),
            "label": f"{node.get_name(accept_no_parent=True)}\n({typename(node)})",
            "type": "group",
            "subtype": typename(subtype),
            "parent": id(node.get_parent_force()[0]) if node.get_parent() else None,
        }
    }


# Style --------------------------------------------------------------------------------


def _with_pastels[T](iterable: Collection[T]):
    return zip(sorted(iterable), generate_pastel_palette(len(iterable)))  # type: ignore


class _Stylesheet:
    _BASE = [
        {
            "selector": "node",
            "style": {
                "content": "data(label)",
                "text-opacity": 0.8,
                "text-valign": "center",
                "text-halign": "center",
                "font-size": "0.5em",
                "background-color": "#BFD7B5",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 0.5,
                "border-width": 1,
                "border-color": "#888888",
                "border-opacity": 0.5,
            },
        },
        {
            "selector": "edge",
            "style": {
                "width": 1,
                "line-color": "#A3C4BC",
                "curve-style": "bezier",
                "target-arrow-shape": "triangle",
                "arrow-scale": 1,
                "target-arrow-color": "#A3C4BC",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 2,
            },
        },
        {
            "selector": 'node[type = "group"]',
            "style": {
                "background-color": "#D3D3D3",
                "font-weight": "bold",
                "font-size": "1.5em",
                "text-valign": "top",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 1.5,
                "text-wrap": "wrap",
                "border-width": 4,
            },
        },
    ]

    def __init__(self):
        self.stylesheet = list(self._BASE)

    def add_node_type(self, node_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'node[type = "{node_type}"]',
                "style": {"background-color": color},
            }
        )

    def add_link_type(self, link_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'edge[type = "{link_type}"]',
                "style": {"line-color": color, "target-arrow-color": color},
            }
        )

    def add_group_type(self, group_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'node[subtype = "{group_type}"]',
                "style": {"background-color": color},
            }
        )


def _Layout(stylesheet: _Stylesheet, elements: list[dict[str, dict]]):
    return html.Div(
        style={
            "position": "fixed",
            "display": "flex",
            "flex-direction": "column",
            "height": "100%",
            "width": "100%",
        },
        children=[
            html.Div(
                className="cy-container",
                style={"flex": "1", "position": "relative"},
                children=[
                    cyto.Cytoscape(
                        id="graph-view",
                        stylesheet=stylesheet.stylesheet,
                        style={
                            "position": "absolute",
                            "width": "100%",
                            "height": "100%",
                            "zIndex": 999,
                        },
                        elements=elements,
                        layout={
                            "name": "fcose",
                            "quality": "proof",
                            "animate": False,
                            "randomize": False,
                            "fit": True,
                            "padding": 50,
                            "nodeDimensionsIncludeLabels": True,
                            "uniformNodeDimensions": False,
                            "packComponents": True,
                            "nodeRepulsion": 1000,
                            "idealEdgeLength": 50,
                            "edgeElasticity": 0.45,
                            "nestingFactor": 0.1,
                            "gravity": 0.25,
                            "numIter": 2500,
                            "tile": True,
                            "tilingPaddingVertical": 10,
                            "tilingPaddingHorizontal": 10,
                            "gravityRangeCompound": 1.5,
                            "gravityCompound": 1.5,
                            "gravityRange": 3.8,
                            "initialEnergyOnIncremental": 0.5,
                            "componentSpacing": 40,
                        },
                    )
                ],
            ),
        ],
    )


# --------------------------------------------------------------------------------------


def interactive_subgraph(
    edges: Iterable[tuple[GraphInterface, GraphInterface, Link]],
    gifs: list[GraphInterface],
    nodes: Iterable[Node],
    height: int | None = None,
):
    links = [link for _, _, link in edges]
    link_types = {typename(link) for link in links}
    gif_types = {typename(gif) for gif in gifs}

    elements = (
        [_gif(gif) for gif in gifs]
        + [_link(*edge) for edge in edges]
        + [_group(node) for node in nodes]
    )

    # Build stylesheet
    stylesheet = _Stylesheet()

    gif_type_colors = list(_with_pastels(gif_types))
    link_type_colors = list(_with_pastels(link_types))
    group_types_colors = [
        (typename(group_type), color) for group_type, color in _GROUP_TYPES.items()
    ]

    for gif_type, color in gif_type_colors:
        stylesheet.add_node_type(gif_type, color)

    for link_type, color in link_type_colors:
        stylesheet.add_link_type(link_type, color)

    for group_type, color in group_types_colors:
        stylesheet.add_group_type(group_type, color)

    # Register the fcose layout
    cyto.load_extra_layouts()
    app = Dash(__name__)
    app.layout = _Layout(stylesheet, elements)

    # Print legend
    console = Console()

    for typegroup, colors in [
        ("GIF", gif_type_colors),
        ("Link", link_type_colors),
        ("Node", group_types_colors),
    ]:
        table = Table(title="Legend")
        table.add_column("Type", style="cyan")
        table.add_column("Color", style="green")
        table.add_column("Name")

        for text, color in colors:
            table.add_row(typegroup, f"[on {color}]    [/]", text)

        console.print(table)

    #
    app.run(jupyter_height=height or 1000)


def interactive_graph(
    G: Graph,
    node_types: tuple[type[Node], ...] | None = None,
    depth: int = 0,
    filter_unconnected: bool = True,
    height: int | None = None,
):
    if node_types is None:
        node_types = (Node,)

    # Build elements
    nodes = GraphFunctions(G).nodes_of_types(node_types)
    if depth > 0:
        nodes = [node for node in nodes if len(node.get_hierarchy()) <= depth]

    gifs = [gif for gif in G.get_gifs() if gif.node in nodes]
    if filter_unconnected:
        gifs = [gif for gif in gifs if len(gif.edges) > 1]

    edges = [edge for edge in G.edges if edge[0] in gifs and edge[1] in gifs]
    return interactive_subgraph(edges, gifs, nodes, height=height)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.kicad.transformer import PCB_Transformer
from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy
from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)

# App --------------------------------------------------------------------------


class App(Module):
    class PowerButton(Module):
        switch = L.f_field(F.Switch(F.Electrical))()
        power_in: F.ElectricPower
        power_switched: F.ElectricPower

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(self.power_in, self.power_switched)

        def __preinit__(self):
            self.power_in.hv.connect_via(self.switch, self.power_switched.hv)
            self.power_in.lv.connect(self.power_switched.lv)
            self.power_in.connect_shallow(self.power_switched)

            self.switch.add(
                F.has_explicit_part.by_supplier(
                    "C318884",
                    pinmap={
                        "1": self.switch.unnamed[0],
                        "2": self.switch.unnamed[0],
                        "3": self.switch.unnamed[1],
                        "4": self.switch.unnamed[1],
                    },
                )
            )

    led: F.PoweredLED
    battery: F.Battery
    power_button: PowerButton

    @L.rt_field
    def transform_pcb(self):
        return F.has_layout_transform(transform_pcb)

    def __preinit__(self) -> None:
        self.led.power.connect_via(self.power_button, self.battery.power)

        # Parametrize
        self.led.led.color.constrain_subset(F.LED.Color.YELLOW)
        self.led.led.brightness.constrain_subset(
            TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
        )

        # TODO: remove when we have a LED picker
        self.led.led.add(F.has_explicit_part.by_supplier("C965802"))
        self.led.led.forward_voltage.alias_is(2.4 * P.V)
        self.led.led.max_brightness.alias_is(435 * P.millicandela)
        self.led.led.max_current.alias_is(20 * P.mA)
        self.led.led.color.alias_is(F.LED.Color.YELLOW)

        # TODO remove when we have a battery picker
        self.battery.add(
            F.has_explicit_part.by_supplier(
                "C5239862",
                pinmap={
                    "1": self.battery.power.lv,
                    "2": self.battery.power.hv,
                },
            )
        )
        self.battery.power.voltage.alias_is(3 * P.V)


# PCB layout etc ---------------------------------------------------------------


def transform_pcb(transformer: PCB_Transformer):
    app = transformer.app
    assert isinstance(app, App)

    # Layout
    Point = F.has_pcb_position.Point
    L = F.has_pcb_position.layer_type

    layout = LayoutTypeHierarchy(
        layouts=[
            LayoutTypeHierarchy.Level(
                mod_type=F.PoweredLED,
                layout=LayoutAbsolute(Point((25, 5, 0, L.TOP_LAYER))),
                children_layout=LayoutTypeHierarchy(
                    layouts=[
                        LayoutTypeHierarchy.Level(
                            mod_type=F.LED, layout=LayoutAbsolute(Point((0, 0)))
                        ),
                        LayoutTypeHierarchy.Level(
                            mod_type=F.Resistor,
                            layout=LayoutAbsolute(Point((-5, 0, 180))),
                        ),
                    ]
                ),
            ),
            LayoutTypeHierarchy.Level(
                mod_type=F.Battery,
                layout=LayoutAbsolute(Point((25, 35, 0, L.TOP_LAYER))),
            ),
            LayoutTypeHierarchy.Level(
                mod_type=F.Switch(F.Electrical),
                layout=LayoutAbsolute(Point((35, 10, 45, L.TOP_LAYER))),
            ),
        ]
    )
    app.add(F.has_pcb_layout_defined(layout))
    app.add(F.has_pcb_position_defined(Point((50, 50, 0, L.NONE))))

    app.add(
        F.has_pcb_routing_strategy_greedy_direct_line(
            F.has_pcb_routing_strategy_greedy_direct_line.Topology.DIRECT
        )
    )

    transformer.set_pcb_outline_complex(
        geometry=transformer.create_rectangular_edgecut(
            width_mm=50,
            height_mm=50,
            origin=(50, 50),
        ),
        remove_existing_outline=True,
    )



import pytest

from atopile import telemetry


@pytest.mark.parametrize(
    ("git_remote",),
    [
        ("https://github.com/atopile/atopile.git",),
        ("git@github.com:atopile/atopile.git",),
    ],
)
def test_commonise_project_url(git_remote):
    assert telemetry.commonise_project_url(git_remote) == "github.com/atopile/atopile"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class PM1006(Module):
    """
    Infrared LED particle sensor module PM1006
    adopts the principle of optical scattering to
    detect the variation trend of particle
    (size between 0.3μm to 10μm) concentration in
    the air. There is an infrared light-emitting
    diode and an optoelectronic sensor built-in
    PM1006, and light rays from the light-emitting
    diode will be reflected when pass through
    the particle. The optoelectronic sensor can
    show the concentration of particle in the air
    by detecting the intensity of reflected light.
    Sensor can output measuring value by pulse
    or UART signal.
    """

    class _pm1006_esphome_config(F.has_esphome_config.impl()):
        update_interval = L.p_field(units=P.s, tolerance_guess=0)

        def get_config(self) -> dict:
            obj = self.obj
            assert isinstance(obj, PM1006), "This is not an PM1006!"

            uart = F.is_esphome_bus.find_connected_bus(obj.data)

            return {
                "sensor": [
                    {
                        "platform": "pm1006",
                        "update_interval": self.update_interval,
                        "uart_id": uart.get_trait(F.is_esphome_bus).get_bus_id(),
                    }
                ]
            }

    esphome_config: _pm1006_esphome_config

    power: F.ElectricPower
    data: F.UART_Base

    # components

    # ---------------------------------------------------------------------
    datasheet = L.f_field(F.has_datasheet_defined)(
        "http://www.jdscompany.co.kr/download.asp?gubun=07&filename=PM1006_LED_PARTICLE_SENSOR_MODULE_SPECIFICATIONS.pdf"
    )
    # ---------------------------------------------------------------------

    def __preinit__(self):
        self.power.voltage.constrain_subset(L.Range.from_center(5 * P.V, 0.2 * P.V))
        self.data.baud.constrain_subset(9600 * P.baud)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class Button(Module):
    unnamed = L.list_field(2, F.Electrical)
    height = L.p_field(
        units=P.mm,
        likely_constrained=False,
        soft_set=L.Range(1 * P.mm, 10 * P.mm),
        tolerance_guess=10 * P.percent,
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.S
    )

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.unnamed[0], self.unnamed[1])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file contains a faebryk sample.
"""

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy

# from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class App(Module):
    led = L.f_field(F.LEDIndicator)(use_mosfet=False)
    mcu: F.RP2040_ReferenceDesign
    usb_power: F.USB_C_PSU_Vertical

    def __preinit__(self) -> None:
        # Parametrize - Need to add picker for LED
        # self.led.led.led.color.constrain_subset(F.LED.Color.YELLOW)
        # self.led.led.led.brightness.constrain_subset(
        #     TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
        # )
        # TODO remove when we have a LED picker
        self.led.led.led.add(F.has_explicit_part.by_supplier("C136118"))

        self.usb_power.power_out.connect(self.mcu.usb.usb_if.buspower)
        self.mcu.rp2040.gpio[25].connect(self.led.logic_in)
        self.mcu.rp2040.pinmux.enable(self.mcu.rp2040.gpio[25])

        self._set_layout()

        # TODO remove when we have a fuse picker
        fuse = self.usb_power.get_first_child_of_type(F.Fuse)
        fuse.add(F.has_explicit_part.by_supplier("C914087"))
        fuse.fuse_type.alias_is(F.Fuse.FuseType.RESETTABLE)
        fuse.response_type.alias_is(F.Fuse.ResponseType.FAST)
        fuse.trip_current.alias_is(1 * P.A)

    def _set_layout(self):
        LT = F.has_pcb_position.layer_type
        Point = F.has_pcb_position.Point
        # set center
        self.add(F.has_pcb_position_defined(Point((50, 50, 0, LT.TOP_LAYER))))

        layout = LayoutTypeHierarchy(
            layouts=[
                LayoutTypeHierarchy.Level(
                    mod_type=type(self.led),
                    layout=LayoutAbsolute(Point((0, 0, 0, LT.NONE))),
                    children_layout=LayoutExtrude((0, -5)),
                    direct_children_only=False,
                ),
                LayoutTypeHierarchy.Level(
                    mod_type=type(self.usb_power),
                    layout=LayoutAbsolute(Point((-20, 0, 0, LT.NONE))),
                    children_layout=LayoutExtrude((0, -5)),
                    direct_children_only=False,
                ),
                LayoutTypeHierarchy.Level(
                    mod_type=type(self.mcu),
                    layout=LayoutAbsolute(Point((30, 0, 0, LT.NONE))),
                ),
            ]
        )
        self.add(F.has_pcb_layout_defined(layout))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Winbond_W25Q16JVUXIQ(F.SPIFlash):
    """
    16Mbit USON-8-EP(2x3) NOR FLASH ROHS
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    explicit_part = L.f_field(F.has_explicit_part.by_mfr)("Winbond", "W25Q16JVUXIQ")

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.qspi.clock.line: ["CLK"],
                self.qspi.chip_select.line: ["CS#", "~{CS}"],
                self.qspi.data[0].line: ["DI(IO0)"],
                self.qspi.data[1].line: ["DO(IO1)"],
                self.power.lv: ["GND", "EP"],
                self.qspi.data[3].line: ["HOLD#orRESET#(IO3)"],
                self.power.hv: ["VCC"],
                self.qspi.data[2].line: ["WP#(IO2)"],
            },
            accept_prefix=True,
            case_sensitive=False,
        )

    @L.rt_field
    def decoupled(self):
        return F.can_be_decoupled_rails(self.power)

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        (self.memory_size <= 16 * P.Mbit).constrain()

        self.power.voltage.constrain_subset(L.Range(2.7 * P.V, 3.6 * P.V))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.core.module import Module
from faebryk.core.parameter import EnumDomain
from faebryk.core.solver.solver import Solver
from faebryk.libs.library import L
from faebryk.libs.sets.sets import EnumSet
from faebryk.libs.smd import SMDSize
from faebryk.libs.util import cast_assert


class has_package_requirements(Module.TraitT.decless()):
    """
    Collection of constraints for package of module.
    """

    size = L.p_field(domain=EnumDomain(SMDSize))

    def __init__(self, *, size: SMDSize | EnumSet[SMDSize] | None = None) -> None:
        super().__init__()

        self._size = size

    def __preinit__(self):
        if self._size is not None:
            self.size.constrain_subset(EnumSet(self._size))

    def get_sizes(self, solver: Solver) -> EnumSet[SMDSize]:
        ss = self.size.get_last_known_deduced_superset(solver)
        return cast_assert(EnumSet, ss)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass

from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.exporters.pcb.layout.layout import Layout
from faebryk.libs.util import find_or, groupby, not_none

logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=True)
class LayoutTypeHierarchy(Layout):
    @dataclass(frozen=True, eq=True)
    class Level:
        mod_type: type[Module] | tuple[type[Module], ...]
        layout: Layout
        children_layout: Layout | None = None
        direct_children_only: bool = True

    layouts: list[Level]

    def apply(self, *node: Node):
        """
        Tip: Make sure at least one parent of node has an absolute position defined
        """
        from faebryk.library.has_footprint import has_footprint

        # Find the layout for each node (isinstance mod_type) and group by matched level
        levels = groupby(
            {
                n: find_or(
                    self.layouts,
                    lambda layout: isinstance(n, not_none(layout).mod_type),
                    default=None,
                )
                for n in node
            }.items(),
            lambda t: t[1],
        )

        logger.debug(f"Applying to {node}")

        for level, nodes_tuple in levels.items():
            nodes = [n for n, _ in nodes_tuple]

            children = {
                c
                for n in nodes
                for c in n.get_children(
                    direct_only=True, types=(Module, ModuleInterface)
                )
            }
            logger.debug(
                f"Level: {level.mod_type if level else None}," f" Children: {children}"
            )

            # No type match, search for children instead
            if level is None:
                self.apply(*children)
                continue

            level.layout.apply(*nodes)

            if not level.children_layout:
                continue

            if not level.direct_children_only:
                children = {
                    c
                    for n in nodes
                    for c in Module.get_children_modules(
                        n,
                        direct_only=False,
                        types=Module,
                        f_filter=lambda c: c.has_trait(has_footprint),
                    )
                }

            level.children_layout.apply(*children)

    def __hash__(self):
        return sum(map(hash, self.layouts))
